# four-Parameter beta Distribution ----------------------------------------------------
#' @title The four-Parameter beta Distribution.
#' @description Density, distribution function, quantile function, random 
#' generation function and parameter estimation function (based on weighted or 
#' unweighted i.i.d. sample) for the 4 Parameter beta distribution 
#' @rdname Beta.ab
#' @name Beta.ab
#' @aliases dBeta.ab pBeta.ab qBeta.ab rBeta.ab eBeta.ab lBeta.ab sBeta.ab
#' @details 
#' \itemize{
#'  \item{Probability density function:}
#'  \deqn{f(x) = \frac{x^{\alpha-1}(1-x)^{\beta-1}} {\mathcal{B}(\alpha,\beta)}}
#'  with \eqn{\alpha} and \eqn{\beta} two shape parameters and \eqn{\mathcal B} beta function
#'  \item{Cumulative distribution function:}
#'  \deqn{F(x) = \frac{\int_{0}^{x} y^{\alpha-1}(1-y)^{\beta-1}dy} {\mathcal{B}(\alpha,\beta)}
#'  =\mathcal{B}(x; \alpha,\beta)} with \eqn{\mathcal B (x; \alpha,\beta)} being incomplete beta function.
#'  \item{Log-likelihood function:}
#'  \deqn{L(\alpha,\beta;X)=\sum_i\left[ (\alpha-1)\ln(x)+(\beta-1)\ln(1-x)-\ln \mathcal{B}(\alpha,\beta) \right]}
#'  \item{Score function vector:}
#'  \deqn{V(\mu,\sigma;X)
#'  =\left( \begin{array}{c}
#'  \frac{\partial L}{\partial \alpha}  \\
#'  \frac{\partial L}{\partial \beta}
#'  \end{array} \right)
#'  =\sum_i
#'  \left( \begin{array}{c}
#'  \psi^{(0)}(\alpha+\beta)-\psi^{(0)}(\alpha)+\ln(x) \\
#'  \psi^{(0)}(\alpha+\beta)-\psi^{(0)}(\beta)+\ln(x) 
#'  \end{array} \right)
#'  }
#'  with \eqn{\psi^{(0)}} being log-gamma function.
#'  \item{Observed information matrix:}
#'  \deqn{\mathcal J (\mu,\sigma;X)=
#'  \left( \begin{array}{cc}
#'  \psi^{(1)}(\alpha)-\psi^{(1)}(\alpha+\beta) & -\psi^{(1)}(\alpha+\beta) \\
#'  -\psi^{(1)}(\alpha+\beta) & \psi^{(1)}(\beta)-\psi^{(1)}(\alpha+\beta) \end{array} \right)
#'  }
#'  with \eqn{\psi^{(1)}} being digamma function.
#' }
#' @param params a list includes all parameters
#' @param x,q vector of quantiles.
#' @param w weights of sample.
#' @param p vector of probabilities.
#' @param n number of observations.
#' @param X sample observations.
#' @param shape1,shape2 shape parameters.
#' @param a,b boundary parameters.
#' @param method parameter estimation method.
#' @param logL logical; if TRUE, lBeta.ab gives log likelihood.
#' @param ... other parameters

#' @return dBeta.ab gives the density; pBeta.ab gives the distribution function;
#' qBeta.ab gives the quantile function; rBeta.ab generates random variables; 
#' eBeta.ab estimate the parameters; sBeta.ab gives observed scorn function 

#' @author Haizhen Wu and A. Jonathan R. Godfrey

#' @examples \donttest{
#' # Parameter estimation
#' n <- 500
#' a <- 1
#' b <- 2
#' shape1 <- 2
#' shape2 <- 5
#' X <- rBeta.ab(n, shape1, shape2, a, b)
#' (est.par <- eBeta.ab(X))
#' 
#' # Histogram and fitted density
#' den.x <- seq(min(X),max(X),length=100)
#' den.y <- dBeta.ab(den.x,params = est.par)
#' hist(X, breaks=10, col="red", probability=TRUE, ylim = c(0,1.1*max(den.y)))
#' lines(den.x, den.y, col="blue", lwd=2)
#' 
#' # Q-Q plot and P-P plot
#' plot(qBeta.ab((1:n-0.5)/n, params=est.par), sort(X), main="Q-Q Plot", 
#' xlab="Theoretical Quantiles", ylab="Sample Quantiles", xlim = c(a,b), ylim = c(a,b))
#' abline(0,1)
#' 
#' plot((1:n-0.5)/n, pBeta.ab(sort(X), params=est.par), main="P-P Plot", 
#' xlab="Theoretical Percentile", ylab="Sample Percentile", xlim = c(0,1), ylim = c(0,1))
#' abline(0,1)
#' 
#' # A weighted parameter estimation example
#' n <- 10
#' par <- list(shape1=2, shape2=5, a= 1, b=2)
#' X <- rBeta.ab(n, params=par)
#' w <- c(0.13, 0.06, 0.16, 0.07, 0.2, 0.01, 0.06, 0.09, 0.1, 0.12)
#' eBeta.ab(X,w) # estimated parameters of weighted sample
#' eBeta.ab(X) # estimated parameters of unweighted sample
#' 
#' # Extracting shape parameters
#' est.par[attr(est.par,"par.type")=="shape"]
#'  
#' # evaluate the performance of the parameter estimation function by simulation
#' eval.estimation(rdist=rBeta.ab,edist=eBeta.ab,n = 1000, rep.num = 1e3, 
#' params = list(shape1=2, shape2=5, a=0, b=1), method ="numerical.MLE")
#' 
#' # evaluate the precision of estimation by Hessian matrix
#' X <- rBeta.ab(1000, shape1, shape2, a, b)
#' (est.par <- eBeta.ab(X))
#' H <- attr(eBeta.ab(X, method = "numerical.MLE"),"nll.hessian")
#' fisher_info <- solve(H)
#' sqrt(diag(fisher_info))
#' 
#' # log-likelihood, score vector and observed information matrix 
#' lBeta.ab(X,param = est.par)
#' lBeta.ab(X,param = est.par, logL=FALSE)
#' sBeta.ab(X,param = est.par)
#' }

#' @rdname Beta.ab
#' @export dBeta.ab
dBeta.ab <-
  function(x, shape1=2, shape2=3, a = 0, b=1, params = list(shape1, shape2, a, b)){
    if(!missing(params)){
      shape1 <- params$shape1
      shape2 <- params$shape2
      a <- params$a
      b <- params$b
    }
    out <- (x>=a & x<=b) * dbeta((x-a)/(b-a),shape1,shape2)/(b-a)
    return(out)
  }

#' @rdname Beta.ab
#' @export pBeta.ab
pBeta.ab <- 
  function(q, shape1=2, shape2=3, a = 0, b=1, params = list(shape1=2, shape2 = 5, a = 0, b = 1)){
    if(!missing(params)){
      shape1 <- params$shape1
      shape2 <- params$shape2
      a <- params$a
      b <- params$b
    }
    out <- pbeta((q-a)/(b-a),shape1,shape2)
    return(out)
  }

#' @rdname Beta.ab
#' @export qBeta.ab
qBeta.ab <- 
  function(p, shape1=2, shape2=3, a = 0, b=1, params = list(shape1=2, shape2 = 5, a = 0, b = 1)){
    if(!missing(params)){
      shape1 <- params$shape1
      shape2 <- params$shape2
      a <- params$a
      b <- params$b
    }
    out <- (b-a)*qbeta(p,shape1,shape2) + a
    return(out)
  }

#' @rdname Beta.ab
#' @export rBeta.ab
rBeta.ab <- 
  function(n, shape1=2, shape2=3, a = 0, b = 1, params = list(shape1, shape2, a, b)){
    if(!missing(params)){
      shape1 <- params$shape1
      shape2 <- params$shape2
      a <- params$a
      b <- params$b
    }
    X <- rbeta(n,shape1,shape2)
    out <- (b-a)*X + a
    return(out)
  }

#' @rdname Beta.ab
#' @export eBeta.ab
eBeta.ab <-     
  function(X,w, method ="numerical.MLE"){
    n <- length(X)
    if(missing(w)){
      w <- rep(1,n)
    } else {
      w <- n*w/sum(w)
    }
    
{
  if(method != "numerical.MLE") warning(paste("method ", method, " is not avaliable, use numerial.MLE instead."))  
  method = "numerical.MLE"  
  
  d <- max(X)-min(X)
  est.par <- wmle(X=X, w=w, distname = "Beta.ab",
                  initial=list(shape1=3,shape2=3,a=min(X)-0.1*d,b=max(X)+0.1*d),
                  lower=list(shape1=0,shape2=0,a=-Inf,b=max(X)),
                  upper=list(shape1=Inf,shape2=Inf,a=min(X),b=Inf))
  est.par.se <- sqrt(diag(solve(attr(est.par,"nll.hessian"))))
} 
    
    attr(est.par,"ob") <- X
    attr(est.par,"weights") <- w
    attr(est.par,"distname") <- "Beta.ab"
    attr(est.par,"method") <- method
    attr(est.par,"par.name") <- c("shape1","shape2","a","b")
    attr(est.par,"par.type") <- c("shape","shape","boundary","boundary")
    attr(est.par,"par.vals") <- c(est.par$shape1, est.par$shape2, est.par$a, est.par$b)
    attr(est.par,"par.s.e") <-  est.par.se  
    
    class(est.par) <- "eDist"
    
    return(est.par)
  }

#' @rdname Beta.ab
#' @export lBeta.ab
## (weighted) (log) likelihood function
lBeta.ab <- 
  function(X, w, shape1=2, shape2 =3, a = 0, b = 1,  params = list(shape1, shape2, a, b), logL = TRUE){
    if(!missing(params)){
      shape1 <- params$shape1
      shape2 <- params$shape2
      a <- params$a
      b <- params$b
    }
    
    n <- length(X)
    if(missing(w)){
      w <- rep(1,n)
    } else {
      w <- n*w/sum(w)
    }
    
    #     ll <- sum(w*((shape1-1)*log(X-a)+(shape2-1)*log(b-X)-log(beta(shape1,shape2))-(shape1+shape2-1)*log(b-a)))
    ll <- sum(w*log(dBeta.ab(x=X,params = params)))
    l <- exp(ll)
    
    if(logL) {return(ll)} else{return(l)}
  }

#' @rdname Beta.ab
#' @export sBeta.ab
## (weighted) score vectors
sBeta.ab <- 
  function(X, w, shape1=2, shape2 =3, a = 0, b = 1,  params = list(shape1, shape2, a, b)){
    if(!missing(params)){
      shape1 <- params$shape1
      shape2 <- params$shape2
      a <- params$a
      b <- params$b
    }
    
    n <- length(X)
    if(missing(w)){
      w <- rep(1,n)
    } else {
      w <- n*w/sum(w)
    }
    
    score1 <- sum(w*(digamma(shape1+shape2)-digamma(shape1)+log(X-a)-log(b-a)))
    score2 <- sum(w*(digamma(shape1+shape2)-digamma(shape2)+log(b-X)-log(b-a)))
    score3 <- sum(w*((shape1+shape2-1)/(b-a)-(shape1-1)/(X-a)))
    score4 <- sum(w*((shape2-1)/(b-X)-(shape1+shape2-1)/(b-a)))
    
    score <- c(score1,score2,score3,score4)
    names(score) <- c("shape1","shape2","a","b")
    return(score)
  }

# #' @rdname Beta.ab
# #' @export iBeta.ab
# ## (weighted) (observed) information matrix
# iBeta.ab <- 
#   function(X, w, shape1=2, shape2 =3, a = 0, b = 1,  params = list(shape1, shape2, a, b)){
#     if(!missing(params)){
#       shape1 <- params$shape1
#       shape2 <- params$shape2
#       a <- params$a
#       b <- params$b
#     }
#     
#     n <- length(X)
#     if(missing(w)){
#       w <- rep(1,n)
#     } else {
#       w <- n*w/sum(w)
#     }
#     
#     info11 <- -sum(w*(trigamma(shape1+shape2)-trigamma(shape1)))
#     info12 <- -sum(w*trigamma(shape1+shape2))
#     info21 <- info12
#     info22 <- -sum(w*(trigamma(shape1+shape2)-trigamma(shape2)))
#     info <- matrix(c(info11,info12,info21,info22), nrow=2,ncol=2)
#     rownames(info) <- colnames(info) <- c("shape1","shape2")
#     return(info)
#   }